import { alias, char, int, mysqlTable, timestamp, varchar } from 'drizzle-orm/mysql-core';
import type { MySql2Database } from 'drizzle-orm/mysql2';
import { aliasedTable, and, asc, desc, eq, like, sql } from 'drizzle-orm';
import { teamMember } from './teamMember.ts';
import { user } from './user.ts';
import { teamDraft } from './teamDraft.ts';
import { death } from '@db/schema/death.ts';
import { report } from '@db/schema/report.ts';
import { reportStatus } from '@db/schema/reportStatus.ts';
import { strikeReason } from '@db/schema/strikeReason.ts';
import { strike } from '@db/schema/strike.ts';

type Database = MySql2Database<{ team: typeof team }>;

export const team = mysqlTable('team', {
  id: int('id').primaryKey().autoincrement(),
  name: varchar('name', { length: 255 }).notNull(),
  color: char('color', { length: 7 }).notNull(),
  lastJoined: timestamp('last_joined', { mode: 'string' })
});

export type AddTeamReq = {
  name: string;
  color?: string;
};

export type EditTeamReq = Partial<Omit<typeof team.$inferSelect, 'id'>> & { id: number };

export type DeleteTeamReq = {
  id: number;
};

export type GetTeamsReq = {
  name?: string | null;
  username?: string | null;
  limit?: number;
};

export type GetTeamsFullReq = {};

export type GetTeamByIdReq = {
  id: number;
};

export type GetTeamByNameReq = {
  name: string;
};

export type GetTeamByUserUuidReq = {
  uuid: string;
};

export async function addTeam(db: Database, values: AddTeamReq) {
  let color = values.color;
  if (!color) {
    const teams = await db.query.team.findMany({
      columns: {
        color: true
      },
      orderBy: [desc(team.id)]
    });

    if (teams.length > 0) {
      for (const team of teams) {
        for (let i = 0; i < teamColors.length; i++) {
          if (teamColors[i] == team.color) {
            color = teamColors[(i + 1) % teamColors.length];
            break;
          }
        }
        if (color) break;
      }
    }

    if (!color) color = teamColors[0];
  }

  const teamIds = await db
    .insert(team)
    .values({
      name: values.name,
      color: color
    })
    .$returningId();

  return {
    id: teamIds[0].id,
    name: values.name,
    color: color
  };
}

export async function editTeam(db: Database, values: EditTeamReq) {
  await db.update(team).set(values).where(eq(team.id, values.id));
}

export async function deleteTeam(db: Database, values: DeleteTeamReq) {
  await db.delete(team).where(eq(team.id, values.id));
}

export async function getTeams(db: Database, values: GetTeamsReq) {
  const memberOneUser = aliasedTable(user, 'member_one_user');
  const memberTwoUser = aliasedTable(user, 'member_two_user');

  return db
    .select({
      id: team.id,
      name: team.name,
      color: team.color,
      lastJoined: team.lastJoined,
      memberOne: {
        id: memberOneUser.id,
        username: teamDraft.memberOneName,
        uuid: memberOneUser.uuid
      },
      memberTwo: {
        id: memberTwoUser.id,
        username: teamDraft.memberTwoName,
        uuid: memberTwoUser.uuid
      }
    })
    .from(team)
    .where(
      and(
        values?.name != null ? like(team.name, `%${values.name}%`) : undefined,
        values?.username != null ? like(teamDraft.memberOneName, `%${values.username}%`) : undefined,
        values?.username != null ? like(teamDraft.memberTwoName, `%${values.username}%`) : undefined
      )
    )
    .orderBy(asc(team.id))
    .innerJoin(teamDraft, eq(team.id, teamDraft.teamId))
    .leftJoin(memberOneUser, eq(teamDraft.memberOneName, memberOneUser.username))
    .leftJoin(memberTwoUser, eq(teamDraft.memberTwoName, memberTwoUser.username))
    .limit(values?.limit ?? 100);
}

export async function getTeamsFull(db: Database, _values: GetTeamsFullReq) {
  const memberOneTeamMember = alias(teamMember, 'member_one_team_member');
  const memberTwoTeamMember = alias(teamMember, 'member_two_team_member');
  const memberOneUser = alias(user, 'member_one_user');
  const memberTwoUser = alias(user, 'member_two_user');
  const memberOneDeath = alias(death, 'member_one_death');
  const memberTwoDeath = alias(death, 'member_two_death');

  const strikeWeightSubquery = db
    .select({
      teamId: team.id,
      strikeWeight: sql<number>`cast(sum(${strikeReason.weight}) as int)`.as('strikeWeight')
    })
    .from(strike)
    .innerJoin(strikeReason, eq(strike.strikeReasonId, strikeReason.id))
    .innerJoin(reportStatus, eq(strike.id, reportStatus.strikeId))
    .innerJoin(report, eq(reportStatus.reportId, report.id))
    .innerJoin(team, eq(report.reportedTeamId, team.id))
    .as('strike_weight_subquery');

  return db
    .select({
      team: team,
      memberOne: memberOneUser,
      memberTwo: memberTwoUser,
      memberOneDeath: memberOneDeath,
      memberTwoDeath: memberTwoDeath,
      strikeWeight: strikeWeightSubquery.strikeWeight
    })
    .from(team)
    .leftJoin(memberOneTeamMember, eq(team.id, memberOneTeamMember.teamId))
    .leftJoin(memberTwoTeamMember, eq(team.id, memberTwoTeamMember.teamId))
    .leftJoin(memberOneUser, eq(memberOneTeamMember.userId, memberOneUser.id))
    .leftJoin(memberTwoUser, eq(memberTwoTeamMember.userId, memberTwoUser.id))
    .leftJoin(memberOneDeath, eq(memberOneUser.id, memberOneDeath.deadUserId))
    .leftJoin(memberTwoDeath, eq(memberTwoUser.id, memberTwoDeath.deadUserId))
    .leftJoin(strikeWeightSubquery, eq(team.id, strikeWeightSubquery.teamId));
}

export async function getTeamById(db: Database, values: GetTeamByIdReq) {
  const teams = await db.select().from(team).where(eq(team.id, values.id));

  return teams[0] ?? null;
}

export async function getTeamByName(db: Database, values: GetTeamByNameReq) {
  const teams = await db.select().from(team).where(eq(team.name, values.name));

  return teams[0] ?? null;
}

export async function getTeamByUserUuid(db: Database, values: GetTeamByUserUuidReq) {
  const teams = await db
    .select({ team })
    .from(team)
    .innerJoin(teamMember, eq(team.id, teamMember.teamId))
    .innerJoin(user, eq(teamMember.userId, user.id))
    .where(eq(user.uuid, values.uuid));

  return teams[0] ?? null;
}

const teamColors = [
  '#cd853f',
  '#ff7f50',
  '#66cdaa',
  '#ffd700',
  '#ba55d3',
  '#d8bfd8',
  '#dc143c',
  '#ff6347',
  '#2e8b57',
  '#daa520',
  '#deb887',
  '#00fa9a',
  '#778899',
  '#ffa500',
  '#e9967a',
  '#b03060',
  '#8b0000',
  '#bdb76b',
  '#663399',
  '#d2b48c',
  '#00ced1',
  '#d2691e',
  '#32cd32',
  '#ff4500',
  '#0000ff',
  '#adff2f',
  '#b22222',
  '#20b2aa',
  '#00bfff',
  '#fa8072',
  '#cd5c5c',
  '#9370db',
  '#00ff00',
  '#006400',
  '#00008b',
  '#f08080',
  '#f4a460',
  '#c71585',
  '#6b8e23',
  '#8a2be2',
  '#b8860b',
  '#ff69b4',
  '#000080',
  '#bc8f8f',
  '#ff1493',
  '#00ffff',
  '#b0e0e6',
  '#ff0000',
  '#00ff7f',
  '#dc143c',
  '#dda0dd',
  '#8b008b',
  '#8fbc8f',
  '#9932cc',
  '#ff00ff',
  '#228b22',
  '#5f9ea0',
  '#3cb371',
  '#b0c4de',
  '#00ff00',
  '#87ceeb',
  '#008b8b',
  '#191970',
  '#90ee90',
  '#ff8c00',
  '#4b0082',
  '#db7093',
  '#cd853f',
  '#7fff00',
  '#c71585',
  '#7f0000',
  '#fa8072',
  '#8b4513',
  '#7b68ee',
  '#dda0dd',
  '#a52a2a',
  '#00ced1',
  '#d2691e',
  '#b8860b',
  '#2f4f4f',
  '#663399',
  '#9370db',
  '#da70d6',
  '#b03060',
  '#bc8f8f',
  '#a0522d',
  '#6495ed',
  '#d8bfd8',
  '#adff2f',
  '#8b0000',
  '#800000',
  '#ffd700',
  '#00fa9a',
  '#4682b4',
  '#808000',
  '#6a5acd',
  '#ff1493',
  '#00ff7f',
  '#40e0d0',
  '#ff4500',
  '#1e90ff',
  '#ff6347'
];
